#version 330
#pragma optionNV(unroll all)

#define SPT_LIGHTCOUNT 2  //light count per type
#define DIR_LIGHTCOUNT 1
#define PT_LIGHTCOUNT 5


#define VSM_MIN_VARIANCE 0.000055
#define VSM_LIGHT_BLEEDING_REDUCTION 0.5 //probably 1.0-x for the values used in literature //see http://http.developer.nvidia.com/GPUGems3/gpugems3_ch08.html light bleeding part... too high will reduce shadow detail, too low will let light bleeding through



//Lights:
struct DirectionalLight
{
	vec3  color;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	sampler2D shadowMap;

	//just used in vertex shader:
	mat4  orthoProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct PointLight
{
	vec3  color;
	vec4  position; //has position & rotation
	float intensity;
	int   decayType;

	//just used in vertex shader:
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};

struct SpotLight
{
	vec3  color;
	vec4  position;
	vec4  direction; //standart direciton is 0.0, 0.0, -1.0,... than it gets rotated
	float intensity;
	float innerConeAngle;
	float outerConeAngle;
	int   decayType;
	sampler2D shadowMap;

	//just used in vertex shader
	mat4  persProjMat;
	mat4  transformMat;
	float zFarPlane;
};




//Material:
struct Material
{
	float   lightFactors[5]; //ambientFactor, diffuseFactor, transparency, specularFactor, shininess
	vec3	ambientColor;
	vec3 	diffuseColor; 
	vec3    transparencyColor;

	//texture
	bool	  hasDiffuseTex;
	sampler2D diffuseTex;
	vec2 	  diffuseTextureScaleUV;
};


//UNIFORMS:
uniform Material mat;

uniform DirectionalLight 	dirLights[ DIR_LIGHTCOUNT ];
uniform PointLight			ptLights[   PT_LIGHTCOUNT ];
uniform SpotLight			sptLights[ SPT_LIGHTCOUNT ];

uniform bool				useShadowMapping;


//INs:
in vec4 fragmentNormal;
in vec4 fragmentEye;
in vec2 fragmentTexCoords;

//calculate the lights in view space:
in vec4 sptLightViewSpacePos[ SPT_LIGHTCOUNT ];
in vec4 sptLightViewSpaceDir[ SPT_LIGHTCOUNT ];
in vec4 ptLightViewSpacePos[  PT_LIGHTCOUNT  ];
in vec4 dirLightViewSpaceDir[ DIR_LIGHTCOUNT ];


//shadow mapping
in vec4 fragmentLightSpt[ SPT_LIGHTCOUNT ];
in float fragmentLightLinearDepthSpt[ SPT_LIGHTCOUNT ];

in vec4 fragmentLightDir[ DIR_LIGHTCOUNT ];
in float fragmentLightLinearDepthDir[ DIR_LIGHTCOUNT ]; //linear depth

//not yet used:
in vec4 fragmentLightPt[ PT_LIGHTCOUNT ];
in float fragmentLightLinearDepthPt[ PT_LIGHTCOUNT ]; //linear depth


//VARIABLES
vec4 N;
vec4 V; //surface position in eye space
vec4 VC; //vector surface camera normalized

//OUT
layout(location = 0, index = 0) out vec4 fragColor;



float linstep(float low, float high, float v){
    return clamp( (v - low)/(high - low), 0.0, 1.0);
}



float calcIntensityAfterDecay( vec4 lightpos, int type, float intensity ){
	if( type == 0 ) return intensity;
	else if( type == 1 ){ //linear
		return intensity / distance( V, lightpos );
	} else if( type == 2 ){ //square
		return intensity / pow( distance( V, lightpos ), 2);
	} 
	//cube
	return intensity / pow( distance( V, lightpos ), 3);
}



float shadowVSM( vec2 moments, float fragLightSpaceDepthLinear ){
	float t = fragLightSpaceDepthLinear;

	float p = float( t <= moments.x );

	float variance = moments.y - (moments.x * moments.x);
	variance = max( VSM_MIN_VARIANCE, variance );
	float d = t - moments.x;
	float pmax = linstep( VSM_LIGHT_BLEEDING_REDUCTION, 1.0, variance/(variance + d*d) );
	return max( p, pmax );
}





vec3 shading(){
	float ambientFactor = mat.lightFactors[0];
	float diffuseFactor = mat.lightFactors[1];

	vec3 diffCol = mat.diffuseColor;
	if( mat.hasDiffuseTex == true ){
	 diffCol = texture( mat.diffuseTex,  fragmentTexCoords * mat.diffuseTextureScaleUV ).rgb;
	}

	vec3 fragcol = diffCol * ambientFactor;
	
	
	//Directional Light:
	for( int i = 0; i < DIR_LIGHTCOUNT; ++i ){
		if( dirLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {

			//shadow mapping:
			float inLight = 1.0;
			if( useShadowMapping == true ){
				vec2 depthLightMoments = textureProj( dirLights[i].shadowMap, fragmentLightDir[i] ).rg;
				inLight = shadowVSM( depthLightMoments, fragmentLightLinearDepthDir[i] );
			}

			if( inLight > 0.008 ){


				vec4 Ldirection = dirLightViewSpaceDir[i];
				float cosLdirN = max( dot( N, -Ldirection ), 0.0 ); //because directional light comes from no point just the normalizes direction is needed to the normal vector
				fragcol += diffCol * dirLights[i].color * dirLights[i].intensity * cosLdirN * diffuseFactor * inLight;
			}
		}
	}


	//Point lights
	for( int i = 0; i < PT_LIGHTCOUNT; ++i ){
		if( ptLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {
			vec4 Lposition = ptLightViewSpacePos[i]; //to eye space
			
			float intensity = calcIntensityAfterDecay( Lposition, ptLights[i].decayType, ptLights[i].intensity );

		
			if( intensity > 0.009 ){
				vec4 L = normalize(Lposition - V);
				float cosLN = max( dot( N, L ), 0.0 );
				fragcol += diffCol * ptLights[i].color * intensity * cosLN * diffuseFactor;
			}
		}
	}


	//Spot lights
	for( int i = 0; i < SPT_LIGHTCOUNT; ++i ){
		if( sptLights[i].color != vec3( 0.0, 0.0, 0.0 ) ) {
			//diffuse part
			vec4 Lposition = sptLightViewSpacePos[i]; //to eye space
			float intensity = calcIntensityAfterDecay( Lposition, sptLights[i].decayType, sptLights[i].intensity );
			if( intensity > 0.009 ){

				//shadow mapping:
				float inLight = 1.0;
				if( useShadowMapping == true ){
					vec2 depthLightMoments = textureProj( sptLights[i].shadowMap, fragmentLightSpt[i] ).rg;
					inLight = shadowVSM( depthLightMoments, fragmentLightLinearDepthSpt[i] );
					//if( inLight == 0.0 ) return vec3( 1.0, 0.0, 0.0 );
				}

				if( inLight > 0.009 ){

					vec4 Ldirection = sptLightViewSpaceDir[i];
					
					vec4 L = normalize(Lposition - V);
					float cosLN = max( dot( N, L ), 0.0 );

					//check if it is in spotlight, outside or in overpass 
					float radLdirVLpos = acos(dot( Ldirection, -L));
					float comparatorInner = sptLights[i].innerConeAngle * 0.5;
					float comparatorOuter = sptLights[i].outerConeAngle * 0.5;
					float insideOutside = clamp( (comparatorOuter - radLdirVLpos) / (comparatorOuter - comparatorInner ), 0.0, 1.0 );
					
					fragcol += diffCol * sptLights[i].color * intensity * cosLN * diffuseFactor * insideOutside * inLight;
				}
			}
		}
	}


	//add ambient term:
	fragcol += mat.ambientColor * mat.lightFactors[0];
	return fragcol;
}


//FOG:
const float  fogStartDistance = 1000.0;
const float  fogMaxDistance   = 1500.0;
const vec4   fogColor		  = vec4( 0.0, 0.0, 0.0, 1.0 );

vec4 useFog( vec4 color ){
	float fogFactor = (fogStartDistance - -V.z)/(fogMaxDistance - fogStartDistance);
	fogFactor = clamp( fogFactor, 0.0, 1.0 );
	return mix(fogColor, color, fogFactor);
}





void main(void)
{
	V = fragmentEye; //surface position in eye space
	VC = normalize( vec4( 0.0, 0.0, 0.0, 0.0 ) - fragmentEye ); //direction surface -> camera
	N = normalize( fragmentNormal ); 

	vec3 fragCol = shading();
	float transparentFactor = mat.lightFactors[2];
	fragColor = vec4(fragCol, transparentFactor);

	//is use fog: SET VALUES!
	//fragColor = useFog(fragColor);
}

